# coding: utf-8
"""

    ElutionalSmootherUtils.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""

class LightIntensityData:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)
