# coding: utf-8
"""
    MctRangeInfo.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""

class MctRangeInfo:
    def __init__(self, conc, ranges, curves):
        self.conc = conc
        self.ranges = ranges
        self.curves = curves
