"""
    KekLib/ContextUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def get_variables_from_context(context, variables_str):
    """
    Get variables from context.
    fully generated by Copy AI.
    """
    variables = variables_str.split(',')
    variables = [v.strip() for v in variables]
    return [context[v] for v in variables]
