"""
    OptJobResultInfo.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

class OptJobResultInfo:
    def __init__(self, fv=None, params=None):
        self.fv = params
        self.params = params
