"""
    AscDscDifference.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt

def show_asc_dsc_difference(js_canvas, debug=True):
    print(show_asc_dsc_difference)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(0, 0, "o")
            fig.tight_layout()
            plt.show()
