"""
    EghAdjuster.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt

class EghAdjuster:
    def __init__(self):
        pass

    def fit(self, fullopt, fv_array, x_list, debug=True):
        pass

    def convert(self, params):
        return params
