"""

    DevelUtils.LegacyImporters.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

def import_legacy_modules():
    import molass_legacy.KekLib, SerialAnalyzer, DataStructure, GuinierAnalyzer, Decomposer, Extrapolation
