"""
    PSD.PsdEstimator.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy.Models.Stochastic.TriporeColumn import TriporeColumn

def compute_psd(model, n, rgs, tRs):
    pore_sv = []
    pore_pv = []

def estimate_psd(model, n):
    pass