# coding: utf-8
"""

    LinearModel.py

"""
import numpy            as np
import scipy.special    as sp

def linear_model(B, x):
    '''Linear function y = m*x + b'''
    # B is a vector of the parameters.
    # x is an array of the current x values.
    # x is in the same format as the x passed to Data or RealData.
    #
    # Return an array in the same format as y passed to Data or RealData.
    return B[0]*x + B[1]
