"""
    Optimizer.OptConstants.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
MIN_NUM_PEAKS = 1
MAX_NUM_PEAKS = 7
MIN_NUM_COMPONENTS = MIN_NUM_PEAKS      # include cases where wihout baseline component
MAX_NUM_COMPONENTS = MAX_NUM_PEAKS + 1
